def generate_protein(model, tokenizer, protein_start = '<gene86>', protein_len = 2000, temperature = 0.8):
    """
        Evaluation step (generating text using the learned model)
        
    """
    
    flag = False
    
    while not flag:
        # Number of characters to generate
        num_generate = protein_len

        # Converting our start string to numbers (vectorizing)
        input_eval = torch.tensor(tokenizer.EncodeAsIds(protein_start))
        input_eval = Variable(input_eval.unsqueeze(0)).cuda()


        # Empty string to store our results
        text_generated = ''
        
        # Low temperatures results in more predictable text.
        # Higher temperatures results in more surprising text.
        # Experiment to find the best setting.
    
    
        # Here batch size == 1
        hidden = None
        for i in range(num_generate):
            flag = False
            predictions, hidden = model(input_eval, [input_eval.size(1)], hidden=hidden)
            predictions = predictions.view(-1).div(temperature).exp()
            check = False in torch.isfinite(predictions)
            if(not check):
                next_protein_int = torch.multinomial(predictions, 1)      
                next_protein_char = tokenizer.DecodeIds(next_protein_int.tolist())
            else:
                print("Inf")
                break
                
            if(next_protein_char == '</gene86>'):
                flag = True
                break

            text_generated += next_protein_char
            if(len(text_generated) >= protein_len):
                flag = True
                break

            input_eval = next_protein_int
            input_eval = Variable(input_eval.unsqueeze(0)).cuda()

    return text_generated


def generate_synth_faa(model_path, model_name, tokenizer_path, num_seq, protein_start, vocab_restrict, temp=0.4, max_len=2119):
    
    hash = hashlib.sha1()
    hash.update(str(time.time()).encode('utf-8'))
    faa_filename = model_path.joinpath('output/' + hash.hexdigest() + "_COG0086.faa")
    faa_filename.parent.mkdir(exist_ok=True)
    
    sp = spm.SentencePieceProcessor()
    sp.Load(str(tokenizer_path))
    
    if(vocab_restrict[0] == True):
        sp.set_vocabulary(vocab_restrict[1])
            
    with open(faa_filename,'a') as fp:
        print(f"Writing to file:{faa_filename}")
        for i in tqdm(range(num_seq)):
            model = torch.load(str(model_path.joinpath(model_name)))
            model.eval()
            sample_seq = Seq(generate_protein(model, sp, protein_start, max_len, temp))
            sample_record = SeqRecord(sample_seq)
            hash.update(str(time.time()).encode('utf-8'))
            sample_record.id = hash.hexdigest() + '_COG0086'
            sample_record.description = f'Synthetic beta\' temp={temp}'
            SeqIO.write(sample_record, fp, "fasta") 



class CharRNN_2L_LSTM(torch.nn.Module):
    def __init__(self):
        """
        In the constructor we instantiate two nn.Linear modules and assign them as
        member variables.
        """
        super(CharRNN_2L_LSTM, self).__init__()
        
        self.encoder = torch.nn.Embedding(100, 64)
        self.lstm_1 = torch.nn.LSTM(64, 128, batch_first=True, num_layers = 2, dropout = 0.3)
        self.dense = torch.nn.Linear(128, 100)
        
        
    def forward(self, seq, len_sorted, hidden = None):
        """
        In the forward function we accept a Tensor of input data and we must return
        a Tensor of output data. We can use Modules defined in the constructor as
        well as arbitrary operators on Tensors.
        """
        if(hidden == None):
            h0 = torch.zeros(2, seq.size(0), 128).cuda().requires_grad_()
            c0 = torch.zeros(2, seq.size(0), 128).cuda().requires_grad_()
        else:
            h0, c0 = hidden
            
        seq = self.encoder(seq)
        seq_pack = torch.nn.utils.rnn.pack_padded_sequence(seq, batch_first=True, lengths=len_sorted).cuda()
        seq_pack_output, (ht, ct) = self.lstm_1(seq_pack, (h0.detach(), c0.detach()))
        output, _ = torch.nn.utils.rnn.pad_packed_sequence(seq_pack_output, batch_first=True)
        
        output = self.dense(output)
        
        return output, (ht, ct)


def train(model, dataloader, optimizer, criterion):
    """
    Returns the train loss
    """
    losses = []
    for seq_input, seq_output, seq_len in dataloader:
        model.train()
        model.zero_grad()
        
        seq_input, seq_output, seq_len = Variable(seq_input), Variable(seq_output), Variable(seq_len)
        seq_input, seq_output, seq_len = seq_input.cuda(), seq_output.cuda(), seq_len.cuda()

        output, _ = model.forward(seq_input, seq_len)
        loss = criterion(output.transpose(1, 2), seq_output)
        loss.backward()
        optimizer.step()
        losses.append(loss.item())
    return losses



def evaluate(model, dataloader, criterion):
    """
    Returns the test/valid loss
    """
    losses = []
    for seq_input, seq_output, seq_len in dataloader:
        model.eval()
        
        seq_input, seq_output, seq_len = Variable(seq_input), Variable(seq_output), Variable(seq_len)
        seq_input, seq_output, seq_len = seq_input.cuda(), seq_output.cuda(), seq_len.cuda()
        
        output, _ = model.forward(seq_input, seq_len)
        loss = criterion(output.transpose(1, 2), seq_output)
        losses.append(loss.item())
        
    return losses 



import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()


setuptools.setup(name='deep_genome',
	version='0.1',
	description='Synthetic Protein Generation',
	url='https://github.com/nitinnairk/deep_genome',
	long_description=long_description,
	long_description_content_type="text/markdown",
	author='Nitin Nair',
	author_email='nitinnair@vt.edu',
	packages=setuptools.find_packages(),
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
		"Operating System :: Unix",
	],
	python_requires='>=3.6',
	zip_safe=False)